local B, L, T = unpack(select(2, ...))

T.LABEL = {
    ['bank'] = L[10001],
    ['reagent'] = L[10002],
    ['void'] = L[10003],
    ['guild'] = L[10004],
    ['player'] = L[10010],
    ['equip'] = L[10018]
}

T.TOOLTIPLABEL = {
    ['bank'] = L[10013],
    ['reagent'] = L[10014],
    ['void'] = L[10016],
    ['guild'] = L[10017],
    ['player'] = L[10015],
    ['equipped'] = L[10012],
    ['auctions'] = AUCTIONS,
    ['mail'] = BUTTON_LAG_MAIL
}

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML bankswatch.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function BW_OnLoad(self)

        ----- REGISTER FOR EVENTS
            self:RegisterEvent('BANKFRAME_OPENED')
            self:RegisterEvent('BANKFRAME_CLOSED')
            self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
            self:RegisterEvent('PLAYERREAGENTBANKSLOTS_CHANGED')
            self:RegisterEvent('VOID_STORAGE_OPEN')
            self:RegisterEvent('VOID_STORAGE_CLOSE')
            self:RegisterEvent('GUILDBANKFRAME_OPENED')
            self:RegisterEvent('GUILDBANK_UPDATE_TABS')
            self:RegisterEvent('PLAYER_ENTERING_WORLD')
            self:RegisterEvent('VARIABLES_LOADED')
            self:RegisterEvent('PLAYER_REGEN_DISABLED')
            self:RegisterEvent('GET_ITEM_INFO_RECEIVED')
            self:RegisterEvent('PLAYER_EQUIPMENT_CHANGED')
            self:RegisterEvent('PLAYERBANKBAGSLOTS_CHANGED')
            self:RegisterEvent('AUCTION_OWNED_LIST_UPDATE')
            self:RegisterEvent('AUCTION_HOUSE_CLOSED')
            self:RegisterEvent('LOADING_SCREEN_DISABLED')
            self:RegisterEvent('MAIL_INBOX_UPDATE')
            self:RegisterEvent('PLAYER_MONEY')

            StackSplitFrame:SetFrameStrata('DIALOG') -- For some reason, this frame is not set as a dialog

            ------ STATIC DIALOG // To get new list name
            StaticPopupDialogs["BWsetminimum"] = {
                text = L[10053],
                button1 = ACCEPT,
                button2 = CANCEL,
                OnShow = function(self)
                    if BWsavedtable['minimumsoundtracker'] then
                        self.text:SetText(self.text:GetText().."\n"..GetMoneyString(BWsavedtable['minimumsoundtracker']))
                    end
                end,
                OnAccept = function (self, data, data2)
                    BWsavedtable['minimumsoundtracker'] = tonumber(MoneyInputFrame_GetCopper(self.moneyInputFrame))
                end,
                hasMoneyInputFrame = 1,
                timeout = 0,
                whileDead = true,
                hideOnEscape = true,
                exclusive = true,
                preferredIndex = 3  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
            }

            StaticPopupDialogs["BWsettreasure"] = {
                text = L[10054],
                button1 = ACCEPT,
                button2 = CANCEL,
                OnShow = function(self)
                    if BWsavedtable['treasuresoundtracker'] then
                        self.text:SetText(self.text:GetText().."\n"..GetMoneyString(BWsavedtable['treasuresoundtracker']))
                    end
                end,
                OnAccept = function (self, data, data2)
                    BWsavedtable['treasuresoundtracker'] = tonumber(MoneyInputFrame_GetCopper(self.moneyInputFrame))
                end,
                hasMoneyInputFrame = 1,
                timeout = 0,
                whileDead = true,
                hideOnEscape = true,
                exclusive = true,
                preferredIndex = 3  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
            }

        DEFAULT_CHAT_FRAME:AddMessage(T.MSGC.DEFAULT..L[10000])

        ------ SLASH COMMANDS
            SLASH_BW1 = "/BW"
            SLASH_BW2 = "/bw"
            SLASH_BW3 = "/Bw"
            SLASH_BW4 = "/bW"
            SLASH_BW5 = "/Bankswatch"
            SLASH_BW6 = "/BanksWatch"
            SLASH_BW7 = "/BANKSWATCH"
            SLASH_BW8 = "/bankswatch"

            SlashCmdList['BW'] = B.Main

            B.CreateFrames()
            table.insert(UISpecialFrames, T.F.mainframe.bank:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.reagent:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.void:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.guild:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.search:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.character:GetName())

    end

    local bag_to_update = nil
    local bag_update = {}
    local money_old = 0
    local money_cumulative = 0
    local difference = 0
    local fontName, fontHeight

    function BW_OnEvent(self, event, ...)
        --B.print(event)

        if event == "VARIABLES_LOADED" then
            B.print(T.MSGC.EVENT.."Variables loaded. Initializing.", 2)
            B.print(T.MSGC.INFO.."Running version: "..GetAddOnMetadata("BanksWatch", "Version"), 1)

            B.C.LoadSettings()

            B.I.Main.Bank.Buttons()
            B.I.Main.Reagent.Buttons()
            B.I.Main.Character.Buttons()
            B.I.Main.Void.Buttons()
            B.I.Main.Guild.Buttons()

            B.print(T.MSGC.INFO.."And now my watch begins...", 1)

        elseif event == "GET_ITEM_INFO_RECEIVED" then
            local itemid = ...
            if T.BW.lookup[itemid] then T.BW.lookup[itemid] = nil end

            if T.BW.lookuprequest and not next(T.BW.lookup) then -- Try to minimize the amount of updates
                T.BW.lookuprequest = false
                B.BankUpdate()
                B.ReagentUpdate()
                B.VoidUpdate()
                B.GuildUpdate()
                B.PlayerUpdate()
                B.PlayerEquipmentUpdate()
            end

        elseif event == "PLAYER_ENTERING_WORLD" then
            -- When loading, BAG_UPDATE triggers, but I only want when player modifies their bank bags
            self:RegisterEvent('PLAYER_LEAVING_WORLD')

            if not T.BW.loaded then -- First time loading character
                T.BW.loaded = true
                self:RegisterEvent('PLAYER_REGEN_DISABLED')
                self:RegisterEvent('MERCHANT_SHOW')

                B.Initialize() -- Load or create character
                B.C.Create() -- Config
                B.SetFrameLocation()

                B.M.Frame.Scale(BWsavedtable['scalemoneytracker'])

                B.PlayerBagsScan()
                B.PlayerEquipmentScan()
                B.PlayerEquipmentUpdate()

                -- Cleanup auctions and add a timer to self delete
                B.CleanMainAuctions()
                B.CleanAltsAuctions()
                B.CleanMainMail()
                B.CleanAltsMail()

                -- Cache items here
                B.CacheMainItems()
                B.CacheAltItems() -- Runs only once

                B.ScrollbarUpdateAuctions()
                B.ScrollbarUpdateMail()
            end

        elseif event == "PLAYER_LEAVING_WORLD" then
            self:UnregisterEvent(event)
            self:UnregisterEvent('BAG_UPDATE')

        elseif event == "LOADING_SCREEN_DISABLED" then -- Register BAG_UPDATE here to avoid spam BAG_UPDATE when zoning
            self:RegisterEvent('BAG_UPDATE')

        elseif event == "PLAYERBANKSLOTS_CHANGED" or event == "PLAYERBANKBAGSLOTS_CHANGED" then
            -- Only for generic bank bag
            local arg = ...

            if arg then -- Sometimes, arg can be nil when getting in combat and set is switched
                if arg <= NUM_BANKGENERIC_SLOTS then -- Generic bank modified
                    B.GenericBankSlotScan(arg)

                elseif arg > NUM_BANKGENERIC_SLOTS then -- Bank bag modified, will trigger BAG_UPDATE
                    bag_to_update = arg - NUM_BANKGENERIC_SLOTS + 4
                end
            end

        elseif event == "PLAYER_EQUIPMENT_CHANGED" then
            B.PlayerEquipmentSlotScan(...)

        elseif event == "BAG_UPDATE" then
            local bag = ...

            if bag == 0 or bag <= NUM_BAG_SLOTS then
                -- Only run when bags are not open since magic doesn't happen unless bags are open.
                if not IsBagOpen(bag) then B.PlayerBagScan(bag) end

            elseif bag_to_update == bag then
                bag_to_update = nil
                B.RefreshBankBags()
                B.BankBagScan(bag)

            elseif T.ISBANKOPEN and T.BW.forceBankScan and (bag > NUM_BAG_SLOTS and bag <= NUM_BAG_SLOTS + NUM_BANKBAGSLOTS) then
                bag_update[tonumber(bag)] = true
            end

        elseif event == "VOID_STORAGE_OPEN" then
            T.ISVOIDOPEN = true
            B.OpeningBank('void')

        elseif event == "VOID_STORAGE_CLOSE" then
            T.ISVOIDOPEN = false
            B.ClosingBank('void')

        elseif event == "BANKFRAME_OPENED" then
            T.ISBANKOPEN = true
            ReagentBankFrame_OnShow(ReagentBankFrame) -- force generate to activate the listener
            B.RefreshBankBags()

            if T.BW.forceBankScan then
                B.BankScan()
            else
                local id
                for id = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do -- open all bank bags
                    if not IsBagOpen(id) then
                        OpenBag(id)
                        CloseBag(id)
                    end
                end
            end

            B.OpeningBank('bank')

        elseif event == "BANKFRAME_CLOSED" then
            T.ISBANKOPEN = false
            B.ClosingBank('bank')

        elseif event == "GUILDBANKFRAME_OPENED" or event == "GUILDBANK_UPDATE_TABS" then
            if event == "GUILDBANKFRAME_OPENED" then self:RegisterEvent('GUILDBANKFRAME_CLOSED') end
            T.ISGUILDOPEN = true
            B.RefreshGuildTab()
            B.OpeningBank('guild')

        elseif event == "GUILDBANKFRAME_CLOSED" then
            self:UnregisterEvent(event)
            T.ISGUILDOPEN = false
            B.ClosingBank('guild')

        elseif event == "PLAYER_REGEN_DISABLED" then
            -- Entering combat, hide any open bags
            B.HideAll()

        elseif event == "PLAYERREAGENTBANKSLOTS_CHANGED" then
            B.ReagentSlotScan(...)

        elseif event == "AUCTION_OWNED_LIST_UPDATE" then
            B.AuctionScan()
            B.CacheMainItems('auctions')
            B.CacheMainItems('player')
            B.ScrollbarUpdateAuctions()

        elseif event == "AUCTION_HOUSE_CLOSED" then
            --B.CacheMainItems('player')

        elseif event == "MERCHANT_SHOW" then
            T.ISMERCHANTOPEN = true
            self:RegisterEvent('MERCHANT_CLOSED')

        elseif event == "MERCHANT_CLOSED" then
            T.ISMERCHANTOPEN = false
            self:UnregisterEvent(event)

        elseif event == "MAIL_INBOX_UPDATE" then
            B.MailScan()
            B.CacheMainItems('mail')
            B.CacheMainItems('player')
            B.ScrollbarUpdateMail()

        elseif event == "PLAYER_MONEY" then
            B.M.Trigger()
        end
    end


    function BW_OnUpdate(self, elapsed)
        -- This is the Blizzard Update, called every computer clock cycle ~millisecond
        if not elapsed then return end -- Otherwise it will infinite loop
        T.BW.total_elapsed = T.BW.total_elapsed + elapsed

        if not T.F.moneyframe.sound and T.BW.total_elapsed > 1 then
            B.M.TreasureTrigger()
        end

        if T.BW.total_elapsed > 2.5 then
            B.M.Reset()
        end

        if T.ISBANKOPEN and T.BW.forceBankScan and next(bag_update) then
            local k, v
            for k, v in pairs(bag_update) do -- try to minimize how many updates to bank bags
                B.BankBagScan(k)
                bag_update[k] = nil
            end
            B.CacheMainItems('bank')
        end
    end

    function BW_ToggleFrame() -- Bindings.xml

        if T.F.mainframe:IsVisible() then
            T.F.mainframe.togglebutton:Show()
        elseif T.F.mainframe.togglebutton:IsVisible() then
            T.F.toggleframe:Hide()
        else
            T.F.toggleframe:Show()
            T.F.mainframe.togglebutton:Hide()
        end
    end

    function BW_ToggleBank(bank) -- Bindings.xml
        local banks = {
            ['bank'] = T.F.mainframe.bankbutton,
            ['reagent'] = T.F.mainframe.reagentbutton,
            ['character'] = T.F.mainframe.characterbutton,
            ['void'] = T.F.mainframe.voidbutton,
            ['guild'] = T.F.mainframe.guildbutton
        }
        T.F.toggleframe:Show()
        T.F.mainframe.togglebutton:Hide()
        T.F.mainframe:Show()
        banks[bank]:Click()
    end

    function BW_ToggleMoneyFrame()
        B.Main("money toggle")
    end

    function B.OpeningBank(bank)

        local banks = {
            ['bank'] = T.F.mainframe.bank,
            ['void'] = T.F.mainframe.void,
            ['guild'] = T.F.mainframe.guild
        }

        T.F.mainframe.characterbutton:SetBackdropBorderColor(0.01, 0.79, 0.15, 1)
        T.F.mainframe.character:SetClampedToScreen(true)
        T.F.mainframe.characterbutton:SetScript("OnClick", function(self, button)

            if not T.F.mainframe.lastbank then 
                B.SetFrameLocation(true) -- force refresh to reclip everything together
            else
                B.AnchorBags(banks[bank])
            end

            if T.F.mainframe.character:IsVisible() then
                T.F.mainframe.character:Hide()
            else
                T.F.mainframe.character:Show()
            end
        end)

        if BWsavedtable['openbank'] and not banks[bank]:IsVisible() then T.F.mainframe.character:Hide(); BW_ToggleBank(bank) end
        if BWsavedtable['openplayer'] and not T.F.mainframe.character:IsVisible() then T.F.mainframe.characterbutton:Click() end
    end

    function B.ClosingBank(bank)

        B.HideAll()
        T.F.mainframe.lastbank = nil
        T.F.mainframe.characterbutton:SetBackdropBorderColor(0, 0, 0, 0)
        T.F.mainframe.characterbutton:SetScript("OnClick", B.I.Main.Frame.Character.Click)
        T.F.mainframe.character:SetClampedToScreen(false)
        B.SetFrameLocation(true) -- force refresh to reclip everything together
    end

--[[//////////////////////////////////////////////////

    MAIN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Initialize()

        hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", B.ContainerFrameItemButton_OnModifiedClick)
        hooksecurefunc("ChatFrame_OnHyperlinkShow", B.ChatFrame_OnHyperlinkShow)
        hooksecurefunc("ChatEdit_InsertLink", B.InsertLink)
        hooksecurefunc("SetItemButtonQuality", B.ButtonListener) -- Magic workaround multiple event triggering

        local _, _, _, enabled, _, _, _ = GetAddOnInfo("Adibags")

        if enabled then T.BW.forceBankScan = true end

        local LibExtraTip = LibStub("LibExtraTip-1")
        LibExtraTip:AddCallback({type = "item", callback = B.TooltipListener})
        LibExtraTip:RegisterTooltip(GameTooltip)

        local player = UnitName('player')
        local server = GetRealmName()
        local faction = UnitFactionGroup('player')

        B.Load(player, server, faction)
    end

    function B.Load(player, server, faction)

        if (player == nil) or (player == UNKNOWNOBJECT) or (player == UNKNOWNBEING) then
            return
        end

        if BWsavedtable and BWsavedtable[server] and BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] then
            B.load(server, player, faction)
        else
            B.print(T.MSGC.EVENT.."New character detected")
            B.template(server, player, faction)
        end

        -- Reference tables for WoW loaded character
        if player == UnitName('player') then
            T['bank'] = B.GetPlayerBank('bank', true)
            T['reagent'] = B.GetPlayerBank('reagent', true)
            T['void'] = B.GetPlayerBank('void', true)
            T['guild'] = B.GetPlayerBank('guild', true)
            T['player'] = B.GetPlayerBank('player', true)
            T['auctions'] = B.GetPlayerBank('auctions', true)
            T['mail'] = B.GetPlayerBank('mail', true)
            T['currency'] = B.GetPlayerBank('currency', true)
            T['currency']['money'] = GetMoney()
        end
    end

    function B.Main(input)

        if input then input = string.lower(input) end
        local input, arg1, val1, val2 = strsplit(" ", input)

        if input == "toggle" then
            if T.F.toggleframe:IsVisible() then
                T.F.toggleframe:Hide()
            else
                T.F.toggleframe:Show()
            end

        elseif input == "debug" then
            BWdebug = not BWdebug
            B.print(T.MSGC.INFO.."Debug:|r "..T.MSGC.BOOL..tostring(BWdebug), 1)

        elseif input == "money" then

            if arg1 then
                if arg1 == "scale" then
                -- verification to make sure nothing silly is happening
                    if not tonumber(val1) then B.print(T.MSGC.ERROR..L[10055], 1); return end

                    val1 = math.max(val1, 75)
                    val1 = math.min(val1, 200)

                    BWsavedtable['scalemoneytracker'] = tonumber(val1)
                    B.print(T.MSGC.INFO.."Scaling:|r "..tostring(BWsavedtable['scalemoneytracker'] / 100), 1)
                    B.M.Frame.Scale(val1)
                    B.Main("money test")

                elseif arg1 == "test" then
                    B.M.Trigger(tonumber(val1 or 100000))

                elseif arg1 == "toggle" then
                    BWsavedtable['enablemoneytracker'] = not BWsavedtable['enablemoneytracker']
                    B.print(T.MSGC.INFO.."Small money tracker:|r "..tostring(BWsavedtable['enablemoneytracker']), 1)
                end
            end
        end
    end

    function B.ButtonListener(button, quality, link)
        -- Workaround to using events which may trigger more than once.
        --B.print(button:GetName())

        if T.BW.loaded then
            if B.IsGuildBank(button) then
                B.GuildSlotScan(GetCurrentGuildBankTab(), button:GetID(), button, link)
            
            elseif B.IsPlayerBag(button) then
                B.PlayerBagSlotScan(button:GetParent():GetID(), button:GetID())

            elseif B.IsBankBag(button) then
                B.BankBagSlotScan(button:GetParent():GetID(), button:GetID())
            
            elseif B.IsReagentBank(button) then
                B.ReagentSlotScan(button:GetID())

            elseif B.IsGenericBank(button) then
                B.GenericBankSlotScan(button:GetID())

            elseif B.IsVoidBank(button) then
                B.VoidSlotScan(tonumber(VoidStorageFrame.page), button:GetID(), button)
            end
        end
    end

    local newstring = {}
    local order = {"player", "equipped", "bank", "reagent", "void", "guild", "mail", "auctions"}
    local r, g, b = T.TOOLTIP.r, T.TOOLTIP.g, T.TOOLTIP.b --0.74, 0.46, 1.00, nice purple...
    local contrast = B.hex(T.TOOLTIPAC.r, T.TOOLTIPAC.g, T.TOOLTIPAC.b)
    function B.TooltipListener(tooltip, itemlink, quantity)
        -- Minimal approach
        local itemid, bank, bag, p, total, server_total, guild, empty_line
        local player = UnitName('player')

        local ignored = {
            [6948] = true, -- Hearthstone
            [110560] = true, -- Garrison Hearthstone
            [140192] = true -- Dalaran Hearthstone
        }

        if itemlink then
            itemid = tonumber(B.GetItemIdFromLink(itemlink))
            if ignored[itemid] then return end

            table.wipe(newstring)
            total = 0
            server_total = 0
            empty_line = false

            -- Main player
            for x = 1, #order do
                bank = order[x]
                bag = T.BW.tooltipcache[player][bank]

                if bag then
                    result = bag[itemid]
                    if result then
                        newstring[#newstring + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                        if bank == "guild" or bank == "mail" then
                            server_total = server_total + result
                        elseif bank ~= "auctions" then
                            total = total + result
                        end
                    end
                end
            end

            if next(newstring) then
                tooltip:AddLine(" ")
                empty_line = true
                server = server_total > 0 and "/|cffffd100"..server_total + total.."|r " or ""
                tooltip:AddDoubleLine(player..contrast.."|r (|cffffffff "..total.." |r"..server..contrast.."|r)", "["..table.concat(newstring, ", ").."]", r, g, b, r, g, b)
            end

            -- Alts
            table.wipe(newstring)
            for p, bank in pairs(T.BW.tooltipcache) do
                if p ~= player then
                    for x = 1, #order do
                        bank = order[x]
                        bag = T.BW.tooltipcache[p][bank]

                        if bag then
                            result = bag[itemid]
                            if result then
                                newstring[#newstring + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                            end
                        end
                    end
                    if next(newstring) then
                        if not empty_line then tooltip:AddLine(" "); empty_line = true end
                        tooltip:AddDoubleLine(p, "["..table.concat(newstring, ", ").."]", r, g, b, r, g, b)
                    end
                    table.wipe(newstring)
                end
            end
        end
    end

    function B.ContainerFrameItemButton_OnModifiedClick(self)

        if IsShiftKeyDown() then
            local bag, item = self:GetParent():GetID(), self:GetID()
            local link = GetContainerItemLink(bag, item)

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then

                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.ChatFrame_OnHyperlinkShow(self, link, text, button)

        if IsShiftKeyDown() and link then
            if string.find(link, 'achievement:*') or string.find(link,'spell:*') then
                return false
            end

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.InsertLink(text)

        if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
            T.F.mainframe.searchbar:SetText(string.find(text, "^|c%x+|H(.+)|h%[.*%]"))
        end
    end

    function B.dropDownMenuItem_OnClick(self)
        local server, player, faction
        local value = self.value

        if not value then return end

        if value == "delete" then
            BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)] = nil

            local player = UnitName('player')
            local server = GetRealmName()
            local faction = UnitFactionGroup('player')
            B.Load(player, server, faction)

            if B.IsCurrentPlayer() then
                B.PlayerBagsScan()
                B.PlayerEquipmentScan()

                B.PlayerUpdate()
                B.PlayerEquipmentUpdate()
            end
            -- Cache items here
            B.CacheMainItems()
            B.CacheAltItems()

            B.ScrollbarUpdateAuctions()
            B.ScrollbarUpdateMail()
            T.F.mainframe.character.auctions:Hide()
            T.F.mainframe.character.mail:Hide()

        elseif value == "movemoneytracker" then
            T.F.moneyframe.anchor:Show()

        elseif value == "enablesoundsmoneytracker" then
            BWsavedtable['enablesoundsmoneytracker'] = not BWsavedtable['enablesoundsmoneytracker']
            return
        elseif value == "enablemoneytracker" then
            BWsavedtable['enablemoneytracker'] = not BWsavedtable['enablemoneytracker']
            return
        elseif value == "minimummoneytracker" then StaticPopup_Show("BWsetminimum")
        elseif value == "treasuremoneytracker" then StaticPopup_Show("BWsettreasure")
        elseif value == "coinssound" then
            BWsavedtable['moneysound'] = 1
            PlaySoundFile(T.SOUNDS[BWsavedtable['moneysound']], "Master")
            return

        elseif value == "retrocoinssound" then
            BWsavedtable['moneysound'] = 2
            PlaySoundFile(T.SOUNDS[BWsavedtable['moneysound']], "Master")
            return

        elseif value == "treasure1sound" then
            BWsavedtable['treasuresound'] = 3
            PlaySoundFile(T.SOUNDS[BWsavedtable['treasuresound']], "Master")
            return

        elseif value == "treasure2sound" then
            BWsavedtable['treasuresound'] = 4
            PlaySoundFile(T.SOUNDS[BWsavedtable['treasuresound']], "Master")
            return
        else
            server, player = value:match"([^.]*).(.*)"
            player, faction = player:match"([^.]*).(.*)"
            faction = faction == "H" and "Horde" or faction == "A" and "Alliance" or faction == "N" and "Neutral"
            B.load(server, player, faction)
        end

        CloseDropDownMenus()
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS (OLDER)

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.BankScan()

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, counter, count, link, bag, slot, slots
        local bank = T['bank']

        if not bank[1] then bank[1] = B.newtable() end
        for x = 1, NUM_BANKGENERIC_SLOTS do
            _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, x)

            if link then
                bank[1][x] = bank[1][x] or B.newtable()
                bank[1][x]['count'] = count
                bank[1][x]['link'] = link
            else
                bank[1][x] = B.deltable(bank[1][x])
            end
        end

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            if not bank[bag] then bank[bag] = B.newtable() end
            
            local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
            local baglink = GetInventoryItemLink("player", inventoryID)

            slots = GetContainerNumSlots(bag)
            bank[bag]['slots'] = slots
            bank[bag]['link'] = baglink

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    bank[bag][x] = bank[bag][x] or B.newtable()
                    bank[bag][x]['link'] = link
                    bank[bag][x]['count'] = count > 1 and count or nil
                else
                    bank[bag][x] = B.deltable(bank[bag][x])
                end
            end
        end

        if B.IsCurrentPlayer() then B.BankUpdate() end
        B.CacheMainItems('bank')
    end

    function B.ReagentScan()

        local x, page, itemID
        local reagent = T['reagent']

        for page = 1, T.REAGENT_PAGES do
            if not reagent[page] then reagent[page] = B.newtable() end

            for x = 1, T.NUM_REAGENT_SLOTS do

                local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, x)

                if link then
                    reagent[page][x] = reagent[page][x] or B.newtable()
                    reagent[page][x]['link'] = link
                    reagent[page][x]['count'] = count > 1 and count or nil
                else
                    reagent[page][x] = B.deltable(reagent[page][x])
                end
            end
        end
    end

    function B.VoidScan()

        if not IsVoidStorageReady() then return end
        local x, page, itemID
        local void = T['void']

        for page = 1, T.VOID_PAGES do
            if not void[page] then void[page] = B.newtable() end

            for x = 1, T.NUM_VOID_SLOTS do
                itemID = GetVoidItemInfo(page, x)

                if itemID then
                    void[page][x] = void[page][x] or B.newtable()
                    void[page][x]['link'] = page == 1 and GetVoidItemHyperlinkString(x) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + x)
                else
                    void[page][x] = B.deltable(void[page][x])
                end
            end
        end
    end

    function B.GuildScan()

        local x, i, id, count, item, button
        local guild = T['guild']
        local page = GetCurrentGuildBankTab()

        if guild[page].access then
            for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                _, count = GetGuildBankItemInfo(page, x)
                if count > 0 then
                    guild[page][x] = guild[page][x] or B.newtable()
                    guild[page][x]['link'] = GetGuildBankItemLink(page, x)
                    guild[page][x]['count'] = count > 1 and count or nil
                else
                    guild[page][x] = B.deltable(guild[page][x])
                end
            end
        end
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.Scan(slot, link, count)

        if link then
            slot = slot or B.newtable()
            slot['link'] = link
            slot['count'] = count and count > 1 and count or nil
        else
            slot = B.deltable(slot)
        end

        return slot
    end

    function B.BankBagScan(bag)

        local x, slots, count, link
        local bank = T['bank']

        B.RefreshBankBag(bank, bag)

        for x = 1, bank[bag]['slots'] do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            bank[bag][x] = B.Scan(bank[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.BankUpdate() end
        B.CacheMainItems('bank')
    end

    function B.GenericBankSlotScan(slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, slot)

        if not B.CompareSlot(bank[1][slot], link, count) then

            bank[1][slot] = B.Scan(bank[1][slot], link, count)
            if B.IsCurrentPlayer() then B.GenericBankSlotUpdate(slot) end
            B.CacheMainItems('bank')
        end
    end

    function B.BankBagSlotScan(bag, slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        B.RefreshBankBag(bank, bag)

        if not B.CompareSlot(bank[bag][slot], link, count) then

            bank[bag][slot] = B.Scan(bank[bag][slot], link, count)
            if B.IsCurrentPlayer() then B.BankSlotUpdate(B.GetSlotID(bag, slot), bag, slot) end
            B.CacheMainItems('bank')
        end
    end

    function B.ReagentSlotScan(slot)

        local reagent = T['reagent']
        local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

        if not reagent[1] then reagent[1] = B.newtable() end

        if not B.CompareSlot(reagent[1][slot], link, count) then

            reagent[1][slot] = B.Scan(reagent[1][slot], link, count)
            if B.IsCurrentPlayer() then B.ReagentSlotUpdate(slot) end
            B.CacheMainItems('reagent')
        end
    end

    function B.VoidSlotScan(page, slot, button)

        local void = T['void']
        local link = nil
        if button.hasItem then link = page == 1 and GetVoidItemHyperlinkString(slot) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + slot) end

        if not void[page] then void[page] = B.newtable() end

        if not B.CompareSlot(void[page][slot], link) then

            void[page][slot] = B.Scan(void[page][slot], link)
            if B.IsCurrentPlayer() then B.VoidSlotUpdate(page, slot) end
            B.CacheMainItems('void')
        end
    end

    function B.GuildSlotScan(page, slot, button, link)

        local guild = T['guild']

        if button.count > 0 then
            if strfind(link, "item:82800:") then -- Caged pet
                link = B.GetGuildPetLink(page, slot)
            end
        end

        if not B.CompareSlot(guild[page][slot], link, button.count) then

            guild[page][slot] = B.Scan(guild[page][slot], link, button.count)
            if B.IsCurrentPlayer() then B.GuildSlotUpdate(page, slot) end
            B.CacheMainItems('guild')
        end
    end

    function B.PlayerBagsScan()

        local x, counter, count, link, bag, slot, slots
        local player = T['player']

        for bag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do

            if not player[bag] then player[bag] = B.newtable() end

            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                player[bag][x] = B.Scan(player[bag][x], link, count)
            end
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
        B.CacheMainItems('player')
    end

    function B.PlayerBagScan(bag)

        local x, slots, count, link
        local player = T['player']

        if not player[bag] then player[bag] = B.newtable() end
            
        slots = GetContainerNumSlots(bag)
        player[bag]['slots'] = slots

        if bag > 0 then
            local inventoryID = ContainerIDToInventoryID(bag)
            local baglink = GetInventoryItemLink("player", inventoryID)
            player[bag]['link'] = baglink
        end

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            player[bag][x] = B.Scan(player[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
        B.CacheMainItems('player')
    end

    function B.PlayerBagSlotScan(bag, slot)

        local player = T['player']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end

        if not B.CompareSlot(player[bag][slot], link, count) then

            player[bag][slot] = B.Scan(player[bag][slot], link, count)
            if B.IsCurrentPlayer() then B.PlayerSlotUpdate(bag, slot) end
            B.CacheMainItems('player')
        end
    end

    function B.PlayerEquipmentScan()

        local inventory, slotId, link
        local player = T['player']

        if not player[-1] then player[-1] = B.newtable() end

        -- Tabard slot is slotId 19?
        for x = 1, #T.INVSLOTS do

            slotId = GetInventorySlotInfo(T.INVSLOTS[x])
            link = GetInventoryItemLink("player", slotId)

            if slotId == 19 then slotId = 18 end

            player[-1][slotId] = B.Scan(player[-1][slotId], link)
        end

        B.CacheMainItems('equipped')
    end

    function B.PlayerEquipmentSlotScan(slot)

        local player = T['player']
        local link = GetInventoryItemLink("player", slot)

        if slot == 19 then slot = 18 end

        if not B.CompareSlot(player[-1][slot], link, nil) then

            player[-1][slot] = B.Scan(player[-1][slot], link)
            if B.IsCurrentPlayer() then B.PlayerEquipmentSlotUpdate(slot) end
            B.CacheMainItems('equipped')
        end
    end

    function B.BankOfflineScan(bag)
        -- Item count is always updated. Check the difference for the bag.
        local slot, count, total
        local bank = T['bank']

        total = 0
        for x = 1, bank[bag]['slots'] do
            slot = bank[bag][x]
            if slot and next(slot) then
                count = (GetItemCount(slot['link'], true) or 0) - (GetItemInfo(slot['link']) or 0)
                B.print(slot['link'])
                B.print(count)
                --[[if count then -- item just changed
                    B.print(slot['link'])
                    B.print(count)
                    B.print(slot['count'])
                end]]
            end
        end
    end

    function B.AuctionScan()

        local x, count, status, index, price, bid, bidAmount
        local auctions = T['auctions']
        local _, totalAuctions = GetNumAuctionItems('owner')

        table.wipe(auctions[1])

        for x = 1, totalAuctions do
            _, _, count, _, _, _, _, bid, _, price, bidAmount, _, _, _, _, status = GetAuctionItemInfo('owner', x)
            if status == 0 then
                index = #auctions[1] + 1
                auctions[1][index] = B.Scan(auctions[1][index], GetAuctionItemLink('owner', x), count)
                auctions[1][index]['time'] = GetTime() + T.TIMEINDEX[GetAuctionItemTimeLeft('owner', x)] -- when auction will actually expire
                auctions[1][index]['price'] = price > 0 and price or bidAmount > 0 and bidAmount or bid > 0 and bid
            end
        end
    end

    function B.MailScan()

        local x, a, index
        local mail = T['mail']
        local totalMoney, totalCOD = 0, 0
        local sender, money, CODAmount, daysLeft, itemCount
        local name, count

        table.wipe(mail[1])

        -- Can only load 50 mail
        for x = 1, GetInboxNumItems() do
            _, _, sender, _, money, CODAmount, daysLeft, itemCount = GetInboxHeaderInfo(x)

            if money > 0 then totalMoney = totalMoney + money end
            if CODAmount > 0 then totalCOD = totalCOD + CODAmount end

            if itemCount then
                index = #mail[1] + 1
                for a = 1, itemCount do
                    name, _, _, count = GetInboxItem(x, a)
                    mail[1][index] = B.Scan(mail[1][index], GetInboxItemLink(x, a), count)
                    mail[1][index]['time'] = GetTime() + (daysLeft*24*60*60)
                    mail[1][index]['money'] = money > 0 and money or CODAmount > 0 and -CODAmount or nil
                    mail[1][index]['sender'] = sender
                end
            end
        end

        mail['money'] = totalMoney
        mail['cod'] = totalCOD
    end

--[[//////////////////////////////////////////////////

    UPDATE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.ScrollbarUpdate()
        local r, g, b, idx, link, hexcolor, itemRarity
        local total_items = #T.BW.items
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.search.scroll)

        FauxScrollFrame_Update(T.F.mainframe.search.scroll, total_items, B.RowsToFit(T.F.mainframe.search.list), T.BUTTON_HEIGHT)
        
        for x = 1, B.RowsToFit(T.F.mainframe.search.list) do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.search.list.buttons[x])
            T.F.mainframe.search.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.search.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.search.list.buttons[x]:Hide()
            T.F.mainframe.search.enddot:Hide()
        end

        for x = 1, B.RowsToFit(T.F.mainframe.search.list) do
            -- Get the appropriate item, which will be x + value
            if (T.F.mainframe.layout == 2 or T.F.mainframe.layout == 4) and total_items > 0 and total_items < B.RowsToFit(T.F.mainframe.search.list) then
                local difference = B.RowsToFit(T.F.mainframe.search.list) - total_items
                buttonnum = x + difference
            else
                buttonnum = x
            end

            idx = x + currentscrollbarvalue

            if T.BW.items[idx] then
                r, g, b, hexcolor = GetItemQualityColor(T.BW.items[idx].quality)

                T.F.mainframe.search.list.buttons[buttonnum].quest = T.BW.items[idx].quest
                T.F.mainframe.search.list.buttons[buttonnum].quality = T.BW.items[idx].quality
                T.F.mainframe.search.list.buttons[buttonnum].icon:SetTexture(T.BW.items[idx].icon)
                T.F.mainframe.search.list.buttons[buttonnum].leftstring:SetText("|c"..hexcolor..T.BW.items[idx].name)

                if T.BW.items[idx].count > 1 then
                    T.F.mainframe.search.list.buttons[buttonnum].text:SetText(T.BW.items[idx].count)
                end

                if T.BW.items[idx].ilvl and T.BW.items[idx].ilvl ~= "" then
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetText(T.BW.items[idx].ilvl)
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetTextColor(r, g, b, 1)
                end

                --loc, num = T.BW.items[idx].loc:match"([^.]*).(.*)"
                if T.BW.items[idx].loc == "alt" then
                    location = T.BW.items[idx].num
                else
                    location = T.BW.items[idx].num == 1 and T.LABEL[T.BW.items[idx].loc] or T.LABEL[T.BW.items[idx].loc].." "..T.BW.items[idx].num
                end
                T.F.mainframe.search.list.buttons[buttonnum].location = {['loc'] = T.BW.items[idx].loc, ['num'] = T.BW.items[idx].num}
                T.F.mainframe.search.list.buttons[buttonnum].rightstring:SetText(location)
                T.F.mainframe.search.list.buttons[buttonnum].link = T.BW.items[idx].link
                T.F.mainframe.search.list.buttons[buttonnum]:Show()
                T.F.mainframe.search.enddot:Show()
                if idx == total_items then
                    T.F.mainframe.search.enddot:Hide()
                end
            end
        end
    end

    function B.ScrollbarUpdateAuctions()

        local time = GetTime()
        local r, g, b, k, v, idx, link, hexcolor, itemRarity
        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice
        local auctions = B.GetPlayerBank('auctions')[1]
        local total_items = #auctions
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.character.auctions.scroll)
        local rows = B.RowsToFit(T.F.mainframe.character.auctions, 28)

        FauxScrollFrame_Update(T.F.mainframe.character.auctions.scroll, total_items, rows, 28)
        for x = 1, rows do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.character.auctions.list.buttons[x])
            T.F.mainframe.character.auctions.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.character.auctions.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.character.auctions.list.buttons[x].rightstring.time:SetVertexColor(0.02, 0.76, 0.31, 1)
            T.F.mainframe.character.auctions.list.buttons[x]:Hide()
            T.F.mainframe.character.auctions.enddot:Hide()
        end

        local total = 0
        for x = 1, #auctions do
            total = total + auctions[x].price
        end

        for x = 1, rows do
            -- Get the appropriate item, which will be x + value
            idx = x + currentscrollbarvalue

            if auctions[idx] then
                if strmatch(auctions[idx].link, "|Hbattlepet:") then
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(auctions[idx].link)
                else
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture = GetItemInfo(auctions[idx].link)
                end
                if name then
                    r, g, b, hexcolor = GetItemQualityColor(quality)

                    T.F.mainframe.character.auctions.list.buttons[x].quality = quality
                    T.F.mainframe.character.auctions.list.buttons[x].name = name
                    T.F.mainframe.character.auctions.list.buttons[x].icon:SetTexture(texture)
                    T.F.mainframe.character.auctions.list.buttons[x].leftstring:SetText("|c"..hexcolor..name.."|r")
                    T.F.mainframe.character.auctions.list.buttons[x].rightstring:SetText(B.ASGSC(auctions[idx].price, false, nil, true, nil, 9))
                    if auctions[idx].count and auctions[idx].count > 1 then
                        T.F.mainframe.character.auctions.list.buttons[x].text:SetText(auctions[idx].count)
                    end

                    if (class == ARMOR or class == ENCHSLOT_WEAPON) and (quality and quality > 1) then -- ILVL for green quality or higher
                        ilvl = B.DetectILVL(link)
                        if ilvl then
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetText(ilvl > 1 and ilvl or nil)
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        else
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetText(iLevel > 1 and iLevel or nil)
                            T.F.mainframe.character.auctions.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        end
                    end

                    for _, v in ipairs(T.TIMEINDEX) do
                        k = T.TIME[v]
                        if (auctions[idx].time - time) <= v then
                            T.F.mainframe.character.auctions.list.buttons[x].rightstring.time:SetVertexColor(k.r, k.g, k.b, k.a)
                            break
                        end
                    end

                    T.F.mainframe.character.auctions.list.buttons[x].link = auctions[idx].link
                    T.F.mainframe.character.auctions.list.buttons[x]:Show()
                    T.F.mainframe.character.auctions.enddot:Show()
                    if idx == total_items then
                        T.F.mainframe.character.auctions.enddot:Hide()
                    end
                end
            end
        end
        T.F.mainframe.character.auctions.tab:SetNormalTexture(GetCoinIcon(total / 100))
        T.F.mainframe.character.auctions.total = total
        T.F.mainframe.character.auctions.totalitems = total_items
    end

    function B.ScrollbarUpdateMail()

        local time, days = GetTime(), 0
        local r, g, b, k, v, idx, link, hexcolor, itemRarity
        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice
        local mail = B.GetPlayerBank('mail')
        local total_items = #mail[1]
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.character.mail.scroll)
        local rows = B.RowsToFit(T.F.mainframe.character.mail, 28)

        FauxScrollFrame_Update(T.F.mainframe.character.mail.scroll, total_items, rows, 28)
        for x = 1, rows do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.character.mail.list.buttons[x])
            T.F.mainframe.character.mail.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.character.mail.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.02, 0.76, 0.31, 1)
            T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText("")
            T.F.mainframe.character.mail.list.buttons[x].rightstring:SetFont(T.F.mainframe.character.mail.list.buttons[x].rightstring:GetFont(), 12)
            T.F.mainframe.character.mail.list.buttons[x]:Hide()
            T.F.mainframe.character.mail.enddot:Hide()
        end

        for x = 1, rows do
            -- Get the appropriate item, which will be x + value
            local mail = mail[1]
            idx = x + currentscrollbarvalue

            if mail[idx] then
                if strmatch(mail[idx].link, "|Hbattlepet:") then
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(mail[idx].link)
                else
                    name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture = GetItemInfo(mail[idx].link)
                end
                
                if name then
                    r, g, b, hexcolor = GetItemQualityColor(quality)

                    T.F.mainframe.character.mail.list.buttons[x].quality = quality
                    T.F.mainframe.character.mail.list.buttons[x].name = name
                    T.F.mainframe.character.mail.list.buttons[x].icon:SetTexture(texture)
                    T.F.mainframe.character.mail.list.buttons[x].leftstring:SetText("|c"..hexcolor..name.."|r")
                    T.F.mainframe.character.mail.list.buttons[x].rightstring.time:Hide()

                    days = (mail[idx].time - time)/24/60/60

                    if mail[idx].money and mail[idx].money < 0 then
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText(B.ASGSC(math.abs(mail[idx].money), false, nil, true, nil, 9))
                        if days < 1 then
                            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.80, 0.16, 0.10, 1)
                        else
                            T.F.mainframe.character.mail.list.buttons[x].rightstring.time:SetVertexColor(0.87, 0.69, 0.14, 1)
                        end
                        T.F.mainframe.character.mail.list.buttons[x].rightstring.time:Show()
                    else
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetFont(T.F.mainframe.character.mail.list.buttons[x].rightstring:GetFont(), 10)
                        T.F.mainframe.character.mail.list.buttons[x].rightstring:SetText(B.GetDays(days))
                    end

                    if mail[idx].count and mail[idx].count > 1 then
                        T.F.mainframe.character.mail.list.buttons[x].text:SetText(mail[idx].count)
                    end

                    if (class == ARMOR or class == ENCHSLOT_WEAPON) and (quality and quality > 1) then -- ILVL for green quality or higher
                        ilvl = B.DetectILVL(link)
                        if ilvl then
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetText(ilvl > 1 and ilvl or nil)
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        else
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetText(iLevel > 1 and iLevel or nil)
                            T.F.mainframe.character.mail.list.buttons[x].ilvl:SetTextColor(r, g, b, 1)
                        end
                    end

                    T.F.mainframe.character.mail.list.buttons[x].link = mail[idx].link
                    T.F.mainframe.character.mail.list.buttons[x]:Show()
                    T.F.mainframe.character.mail.enddot:Show()
                    if idx == total_items then
                        T.F.mainframe.character.mail.enddot:Hide()
                    end
                end
            end
        end

        T.F.mainframe.character.mail.tab:SetNormalTexture(B.GetMailIcon(total_items))
        T.F.mainframe.character.mail.totalitems = total_items
        T.F.mainframe.character.mail.money = mail['money'] or 0
        T.F.mainframe.character.mail.cod = mail['cod'] or 0
    end

    function B.BankUpdate()

        local page, total, counter, quality, r, g, b
        local bank = B.GetPlayerBank('bank')

        B.ValidateBankBags(bank)

        counter = 0
        for page = 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
            if bank[page] then

                if page > 1 then -- Ignore generic bank
                    B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, bank[page])
                end
            
                B.UpdatePage(T.F.mainframe.bank, bank[page]['slots'], bank[page], counter)
                counter = counter + bank[page]['slots']
            
            elseif page >= NUM_BAG_SLOTS + 1 then
                B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, {})
            end
        end
    end

    function B.PlayerUpdate()

        local page, total, counter, quality, r, g, b
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then

                B.UpdatePage(T.F.mainframe.character, player[page]['slots'], player[page], counter)
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.PlayerSlotUpdate(bag, slot)

        local page, counter, item
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then
                if page == bag then
                    
                    B.ResetButton(T.F.mainframe.character.buttons[counter + slot])
                    if player[bag][slot] then

                        item = B.GetItem(player[bag][slot].link)
                        if next(item) then B.UpdateButton(T.F.mainframe.character.buttons[counter + slot], item, player[bag][slot]) end
                    end
                end
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.GenericBankSlotUpdate(slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[slot], bank[1][slot])
    end

    function B.BankSlotUpdate(button, page, slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[button], bank[page][slot])
    end

    function B.PlayerEquipmentUpdate()
        local player = B.GetPlayerBank('player')
        B.UpdatePage(T.F.mainframe.character.equipment, 18, player[-1])
    end

    function B.PlayerEquipmentSlotUpdate(slot)
        local player = B.GetPlayerBank('player')
        B.UpdateSlot(T.F.mainframe.character.equipment.buttons[slot], player[-1][slot])
    end

    function B.ReagentSlotUpdate(slot)
        local reagent = B.GetPlayerBank('reagent')
        B.UpdateSlot(T.F.mainframe.reagent[1].buttons[slot], reagent[1][slot])
    end

    function B.VoidSlotUpdate(page, slot)
        local void = B.GetPlayerBank('void')
        B.UpdateSlot(T.F.mainframe.void[page].buttons[slot], void[page][slot])
    end

    function B.GuildSlotUpdate(page, slot)
        local guild = B.GetPlayerBank('guild')
        B.UpdateSlot(T.F.mainframe.guild[page].buttons[slot], guild[page][slot])       
    end

    function B.ReagentUpdate()
        local page, reagent = _, B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            B.UpdatePage(T.F.mainframe.reagent[page], T.NUM_REAGENT_SLOTS, reagent[page])
        end
    end

    function B.VoidUpdate()
        local page, void = _, B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do
            B.UpdatePage(T.F.mainframe.void[page], T.NUM_VOID_SLOTS, void[page])
        end
    end

    function B.GuildUpdate()

        local page, guild = _, B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do
            if guild[page] then

                if guild[page].icon then
                    T.F.mainframe.guild[page].tab:SetNormalTexture(guild[page].icon)
                    T.F.mainframe.guild[page].tab.name = guild[page].name
                    T.F.mainframe.guild[page].tab:Show()
                else
                    T.F.mainframe.guild[page].tab:Hide()
                end

                if guild[page].access then -- Validate user has access to view tab
                    T.F.mainframe.guild[page].tab:Enable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(false)
                    B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, guild[page])
                else
                    T.F.mainframe.guild[page].tab:Disable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(true)
                end
            else
                T.F.mainframe.guild[page].tab:Hide()
                B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, nil)
            end
         end
    end

    function B.RefreshGuildTab()

        local x, name, icon, isViewable
        local guild = T['guild']
        local tabs = GetNumGuildBankTabs()

        if IsInGuild() then guild['name'] = GetGuildInfo('player') end

        for x = 1, MAX_GUILDBANK_TABS do
            if x <= tabs then
                if not guild[x] then guild[x] = B.newtable() end
                name, icon, isViewable = GetGuildBankTabInfo(x)
                guild[x]['icon'] = icon
                guild[x]['name'] = name
                guild[x]['access'] = isViewable

                T.F.mainframe.guild[x].tab:SetNormalTexture(icon)
                T.F.mainframe.guild[x].tab.name = name
                T.F.mainframe.guild[x].tab:Enable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(false)
                T.F.mainframe.guild[x].tab:Show()
            else
                guild[x] = B.deltable(guild[x])
                T.F.mainframe.guild[x].tab:Disable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(true)
                T.F.mainframe.guild[x].tab:Hide()
            end
        end
    end

    function B.RefreshBankBags()
        local total, page
        local bank = T['bank']

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            B.RefreshBankBag(bank, bag)

            if bag > 1 then -- Ignore generic bank
                B.UpdateBagTab(T.F.mainframe.bank.bags[bag - 4].tab, bank[bag])
            end
        end

        B.ValidateBankBags(bank)
    end

    function B.RefreshBankBag(bank, bag)

        if not bank[bag] then bank[bag] = B.newtable() end
            
        local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
        local baglink = GetInventoryItemLink("player", inventoryID)
        local slots = GetContainerNumSlots(bag)

        bank[bag]['slots'] = slots
        bank[bag]['link'] = baglink

        B.ValidateBankBags(bank)
    end

    function B.RefreshPlayerBag(player, bag)
        local slots

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end
    end

    function B.UpdateBagTab(tab, data)
        local quality, r, g, b

        if data.link then
            quality = B.GetQualityFromLink(data.link)
            r, g, b = GetItemQualityColor(quality)

            tab:SetBackdropBorderColor(r, g, b, 1)
            tab.link = data.link
            tab.quality = quality
            tab:SetNormalTexture(GetItemIcon(data.link))
            tab:Show()
        else
            tab:Hide()
        end
    end

    function B.UpdatePage(page, maxslots, data, optstart)
        local x, item
        optstart = optstart or 0

        for x = 1, maxslots do
            B.UpdateSlot(page.buttons[x + optstart], data and data[x] or nil)
        end
    end

    function B.UpdateSlot(button, data)

        if button then
            B.ResetButton(button)

            if data then
                item = B.GetItem(data.link)
                if next(item) then B.UpdateButton(button, item, data) end
            end
        end
    end

    local item = {}
    function B.GetItem(itemlink)

        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice

        if strmatch(itemlink, "|Hbattlepet:") then
            name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(itemlink)
        elseif strmatch(itemlink, "|Hkeystone:") then
            name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetKeystoneInfo(itemlink)
        else
            name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(itemlink)
        end

        -- Reuse the same table to avoid generating garbage
        item[1] = name
        item[2] = link
        item[3] = quality
        item[4] = iLevel
        item[5] = reqLevel
        item[6] = class
        item[7] = subclass
        item[8] = maxStack
        item[9] = equipSlot
        item[10] = texture
        item[11] = vendorPrice

        if not next(item) and B.GetItemIdFromLink(itemlink) then -- Item is not cached yet
            T.BW.lookuprequest = true
            T.BW.lookup[tonumber(B.GetItemIdFromLink(itemlink))] = true
        end

        return item
    end

    function B.UpdateButton(button, item, data)
        local r, g, b, ilvl

        r, g, b = GetItemQualityColor(item[3])

        button.name = item[1]
        button.count = data.count or 1
        button.quality = item[3]
        button.link = item[2]
        button.class = item[6]
        button.subclass = item[7]
        button.equipslot = item[9]
        button.sell = item[11] and item[11] > 0 and item[11] or nil

        button.icon:SetTexture(item[10])
        button.icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        
        if (data.count or 1) > 1 then button.text:SetText(data.count) end
        if item[3] > 1 then button:SetBackdropBorderColor(r, g, b, 1) end

        if item[6] == BATTLE_PET_SOURCE_2 then -- QUEST
            button:SetBackdropBorderColor(1, 1, 0, 1)
            button.quest = true

        elseif (item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON) and (button.quality and button.quality > 1) then -- ILVL for green quality or higher
            ilvl = B.DetectILVL(item[2])
            if ilvl then
                button.ilvl:SetText(ilvl > 1 and ilvl or nil)
                button.ilvl:SetTextColor(r, g, b, 1)
            else
                button.ilvl:SetText(item[4] > 1 and item[4] or nil)
                button.ilvl:SetTextColor(r, g, b, 1)
            end
        end
    end

    function B.ResetButton(button)

        button.icon:SetTexture("")
        button:SetBackdropBorderColor(0, 0, 0, 1)
        button.link = nil
        button.quality = nil
        button.quest = nil
        button.name = nil
        button.count = nil
        button.text:SetText("")
        button.ilvl:SetText("")
    end

--[[//////////////////////////////////////////////////

    SEARCH FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Search(text)

        if text == "" or text == SEARCH then text = false end

        T.F.mainframe.search.scroll.ScrollBar:SetValue(0)

        B.deltable(T.BW.items)
        T.BW.items = B.newsearchtable()

        B.BankSearch(text)
        B.ReagentSearch(text)
        B.VoidSearch(text)
        B.GuildSearch(text)
        B.PlayerSearch(text)
        B.PlayerEquipmentSearch(text)
        B.SearchAlts(text)

        B.ScrollbarUpdate()

        if text and T.F.mainframe.countbutton.count > 0 then
            T.F.mainframe.countbutton.labelcount:SetText("|cffffffff"..T.F.mainframe.countbutton.count.."|r")
            T.F.mainframe.countbutton:Show()
            if not T.F.mainframe.search:IsVisible() and not T.F.mainframe.lastbank then
                T.F.mainframe.countbutton:Click()
            end
        else
            T.F.mainframe.countbutton:Hide()
        end
    end

    function B.BankSearch(text)

        local x, page, counter, button, buttons, name
        local bank = B.GetPlayerBank('bank')

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if page > 1 then -- ignore generic bank
                    local tab = page - 4

                    if text then T.F.mainframe.bank.bags[tab].tab.search:Show()
                    else T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.bank.buttons

                for x = 1, bank[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "bank."..(page == 1 and page or page - 4)
                            B.SearchSlot(button)
                            if page > 1 then local tab = page - 4; T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerSearch(text)

        local x, page, counter, button, buttons, name
        local player = B.GetPlayerBank('player')

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    local tab = page

                    if text then T.F.mainframe.character.bags[tab].tab.search:Show()
                    else T.F.mainframe.character.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "player."..(page == 0 and 1 or page + 1)
                            B.SearchSlot(button)
                            if page > 0 then local tab = page; T.F.mainframe.character.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerArtifactSearch()

        local x, page, counter, button, buttons, name
        local player = T['player']

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    T.F.mainframe.character.bags[page].tab.search:Show()
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name then
                        if IsArtifactPowerItem(B.GetItemIdFromLink(button.link)) then
                            if not T.F.mainframe.character.artifact.total then T.F.mainframe.character.artifact.total = B.DetectAP(B.GetBag(button)) end
                            if page > 0 then T.F.mainframe.character.bags[page].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerEquipmentSearch(text)

        local x, button, buttons, name
        local player = B.GetPlayerBank('player')

        if not player[-1] then return end

        buttons = T.F.mainframe.character.equipment.buttons
        for x = 1, T.EQUIPMENT_SLOTS do
            button = buttons[x]

            if button.name and text then
                if B.find(text, button) then
                    button.loc = "equip.1"
                    B.SearchSlot(button)
                else
                    button.search:Show()
                end
            else
                button.search:Hide()
            end

        end
    end

    function B.BankBagSearch(bag)

        local bank = B.GetPlayerBank('bank')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if not bag or page == bag then
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.PlayerBagSearch(bag)

        local player = B.GetPlayerBank('player')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if not bag or page == bag then
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.ReagentSearch(text)

        local x, page, button, buttons
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            buttons = T.F.mainframe.reagent[page].buttons
            for x = 1, T.NUM_REAGENT_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "reagent."..page
                        B.SearchSlot(button)
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.VoidSearch(text)

        local x, page, button, buttons
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do

            if text then T.F.mainframe.void[page].tab.search:Show() 
            else T.F.mainframe.void[page].tab.search:Hide() end

            buttons = T.F.mainframe.void[page].buttons

            for x = 1, T.NUM_VOID_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "void."..page
                        B.SearchSlot(button)
                        T.F.mainframe.void[page].tab.search:Hide()
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.GuildSearch(text)

        local x, page, button, buttons
        local guild = B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do

            if text then T.F.mainframe.guild[page].tab.search:Show()
            else T.F.mainframe.guild[page].tab.search:Hide() end

            if guild[page] and guild[page].access then

                buttons = T.F.mainframe.guild[page].buttons

                for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do
                    button = buttons[x]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "guild."..page
                            B.SearchSlot(button)
                            T.F.mainframe.guild[page].tab.search:Hide()
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                end
            end
        end
    end

    function B.SearchAlts(text)
        local p, b, player, faction, item, link, info, temp, bags, bank
        local player, faction, loc, num, name, link, item
        -- Only same server and faction as selected player
        if not text then return end

        for p, b in pairs(BWsavedtable[T.ACTIVE_TABLE.server]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(T.ACTIVE_TABLE.faction, 1, 1) and player ~= T.ACTIVE_TABLE.player then
                for _, bank in pairs(b) do-- banks, bags
                    for _, bags in pairs(bank) do -- bag index, bags
                        if type(bags) == "table" then
                            for _, item in pairs(bags) do
                                if type(item) == "table" then

                                    link = item['link']
                                    info = strmatch(link, "|Hbattlepet:") and {B.GetPetInfo(link)} or {GetItemInfo(link)}

                                    temp = B.newtable()
                                    temp.name = info[1]
                                    temp.quality = info[3]
                                    temp.link = link
                                    temp.class = info[6]
                                    temp.subclass = info[7]
                                    temp.equipslot = info[9]

                                    if B.find(text, temp) then
                                        B.SearchAltSlot(player, info, item['count'] or 1)
                                    end

                                    B.deltable(temp)
                                end
                            end
                        end
                    end
                end
            end
        end
    end

    function B.SearchSlot(button)

        button.search:Hide()

        loc, num = button.loc:match"([^.]*).(.*)"
        num = tonumber(num)

        result = T.BW.items(loc, num, button.name) -- use metatable __call to group items
        if not next(result) then -- New entry
            result['loc'] = loc
            result['num'] = num
            result['name'] = button.name
            result['link'] = button.link
            result['quest'] = button.quest
            result['quality'] = button.quality
            result['ilvl'] = button.ilvl:GetText()
            result['icon'] = button.icon:GetTexture()
            result['count'] = 0
        end

        result['count'] = result['count'] + button.count
        T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + button.count
    end

    function B.SearchAltSlot(alt, item, count)

        result = T.BW.items("alt", alt, item[1]) -- use metatable __call to group items
        if not next(result) then -- New entry
            result['loc'] = "alt"
            result['num'] = alt
            result['name'] = item[1]
            result['link'] = item[2]
            result['quality'] = item[3]
            result['quest'] = item[6] == BATTLE_PET_SOURCE_2
            result['ilvl'] = item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON and item[4] or nil
            result['icon'] = item[10]
            result['count'] = 0
        end

        result['count'] = result['count'] + count
    end

--[[//////////////////////////////////////////////////

    TOOLTIP CACHE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.CacheItems(cache, itemid, count)
        if not itemid then return end -- keystone don't have itemids

        itemid = tonumber(itemid)

        if not cache[itemid] then
            cache[itemid] = 0
        end

        cache[itemid] = cache[itemid] + count
    end

    function B._CacheItems(player, location, bank)
        local bags, item, b

        if not bank then return end

        if not T.BW.tooltipcache[player][location] then
            T.BW.tooltipcache[player][location] = B.newtable()
        end

        table.wipe(T.BW.tooltipcache[player][location])

        if location == "equipped" then
            for _, item in pairs(bank) do
                if type(item) == "table" then

                    B.CacheItems(T.BW.tooltipcache[player][location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                end
            end
        else
            for b, bags in pairs(bank) do
                if (location == "player" and b > -1) or location ~= "player" then
                    if type(bags) == "table" then
                        for _, item in pairs(bags) do
                            if type(item) == "table" then

                                B.CacheItems(T.BW.tooltipcache[player][location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                            end
                        end
                    end
                end
            end
        end
    end

    function B.CacheMainItems(bank)
        -- Cache everything from active player into a workable table by itemid - count only.
        local banks, selected
        local player = UnitName('player')

        T.BW.tooltipcache[player] =  T.BW.tooltipcache[player] or B.newtable()

        if not bank then -- Cache everything
            local banks = {"bank", "reagent", "void", "guild", "player", "equipped", "auctions", "mail"}
            for bank = 1, #banks do
                selected = banks[bank] == "equipped" and T['player'][-1] or T[banks[bank]]
                B._CacheItems(player, banks[bank], selected)
            end
        else
            selected = bank == "equipped" and T['player'][-1] or T[bank]
            B._CacheItems(player, bank, selected)
        end
    end

    function B.CacheAltItems() -- Run only once
        local player, faction, p, b, banks, bank, selected

        for p, b in pairs(BWsavedtable[GetRealmName()]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(UnitFactionGroup('player'), 1, 1) and player ~= UnitName('player') then
                T.BW.tooltipcache[player] = B.newtable()

                banks = {"bank", "reagent", "void", "guild", "player", "equipped", "auctions", "mail"}
                for bank = 1, #banks do

                    selected = banks[bank] == "equipped" and b['player'][-1] or b[banks[bank]]
                    B._CacheItems(player, banks[bank], selected)
                end
            end
        end
    end

--[[//////////////////////////////////////////////////

    CLEAN UP CREW

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.CleanMainAuctions()

        local x
        local auctions = T['auctions']
        local time = GetTime()

        for x = #auctions[1], 1, -1 do
            if auctions[1][x] then
                if auctions[1][x]['time'] <= time then -- expired
                    table.remove(auctions[1][x], x)
                end
            end
        end
    end

    function B.CleanAltsAuctions()

        local player, faction, p, b, auctions, time, x

        for p, b in pairs(BWsavedtable[GetRealmName()]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(UnitFactionGroup('player'), 1, 1) and player ~= UnitName('player') then
                auctions = b['auctions']
                time = GetTime()

                if auctions then
                    for x = #auctions[1], 1, -1 do
                        if auctions[1][x] then
                            if auctions[1][x]['time'] <= time then -- expired
                                table.remove(auctions[1], x)
                            end
                        end
                    end
                end
            end
        end
    end

    function B.CleanMainMail()

        local x
        local mail = T['mail']
        local time = GetTime()

        for x = #mail[1], 1, -1 do
            if mail[1][x] then
                if mail[1][x]['time'] <= time then -- expired
                    table.remove(mail[1][x], x)
                end
            end
        end
    end

    function B.CleanAltsMail()

        local player, faction, p, b, mail, time, x

        for p, b in pairs(BWsavedtable[GetRealmName()]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(UnitFactionGroup('player'), 1, 1) and player ~= UnitName('player') then
                mail = b['mail']
                time = GetTime()

                if mail then
                    for x = #mail[1], 1, -1 do
                        if mail[1][x] then
                            if mail[1][x]['time'] <= time then -- expired
                                table.remove(mail[1], x)
                            end
                        end
                    end
                end
            end
        end
    end
